@@ -2,7 +2,7 @@
  * OCaml bindings for libmad
  *
  * Copyright 2001 Bertrand Petit
- * Copyright 2003 Savonet team
+ * Copyright 2003-2006 Savonet team
  *
  * This file is part of Ocaml-mad.
  * The libmad's stubs are based on madlld (see README for details).
@@ -24,16 +24,15 @@
  * @author Samuel Mimram
  */
 
-/* $Id: mad_stubs.c,v 1.28 2005/03/03 11:09:14 smimram Exp $ */
+/* $Id: mad_stubs.c 2395 2006-04-10 09:11:58Z smimram $ */
 
-#define CAML_NAME_SPACE
 #include <caml/alloc.h>
 #include <caml/callback.h>
+#include <caml/custom.h>
 #include <caml/fail.h>
 #include <caml/memory.h>
 #include <caml/misc.h>
 #include <caml/mlvalues.h>
-#include <caml/custom.h>
 #include <caml/signals.h>
 
 #include <stdio.h>
@@ -53,7 +52,8 @@ struct madfile__t
   struct mad_synth synth;
   mad_timer_t timer;
   FILE *fd;
-  char *buf;
+  value read_func;
+  unsigned char *buf;
 };
 
 typedef struct madfile__t madfile_t;
@@ -99,14 +99,13 @@ static unsigned short ushort_of_madfixed(mad_fixed_t sample)
   return ((unsigned short)sample);
 }
 
-static inline madfile_t* madfile_of_block(value block)
-{
-  return (madfile_t*)Field(block, 1);
-}
+#define Madfile_val(v) (*((madfile_t**)Data_custom_val(v)))
 
-static void finalize_madfile_t(value madf)
+static void finalize_madfile(value madf)
 {
-  madfile_t *mf = madfile_of_block(madf);
+  madfile_t *mf = Madfile_val(madf);
+  if (mf->read_func)
+    caml_remove_global_root(&mf->read_func);
   mad_synth_finish(&mf->synth);
   mad_frame_finish(&mf->frame);
   mad_stream_finish(&mf->stream);
@@ -114,34 +113,71 @@ static void finalize_madfile_t(value madf)
   free(mf);
 }
 
+static struct custom_operations madfile_ops =
+{
+  "ocaml_mad_madfile",
+  finalize_madfile,
+  custom_compare_default,
+  custom_hash_default,
+  custom_serialize_default,
+  custom_deserialize_default
+};
+
+static inline madfile_t* create_mf()
+{
+  madfile_t *mf;
+
+  mf = malloc(sizeof(madfile_t));
+  mad_stream_init(&mf->stream);
+  mad_frame_init(&mf->frame);
+  mad_synth_init(&mf->synth);
+  mad_timer_reset(&mf->timer);
+
+  return(mf);
+}
+
 CAMLprim value ocaml_mad_openfile(value file)
 {
   CAMLparam1(file);
   CAMLlocal1(block);
   madfile_t *mf;
-  mf = (madfile_t*)malloc(sizeof(madfile_t));
-  mad_stream_init(&mf->stream);
-  mad_frame_init(&mf->frame);
-  mad_synth_init(&mf->synth);
-  mad_timer_reset(&mf->timer);
-  mf->fd = fopen(String_val(file), "r");
-  if (!mf->fd) {
-    free(mf);
-    caml_raise_with_arg(*caml_named_value("mad_exn_openfile_error"),
-		   caml_copy_string(strerror(errno)));
-  }
-  mf->buf = (char*)malloc(BUFFER_SIZE);
 
-  block = caml_alloc_final(2, finalize_madfile_t, sizeof(madfile_t), BUFFER_SIZE + 1024);
-  Store_field(block, 1, (value)mf);
+  mf = create_mf();
+  mf->read_func = (value)NULL;
+  mf->fd = fopen(String_val(file), "r");
+  if (!mf->fd)
+    caml_raise_with_arg(*caml_named_value("mad_exn_openfile_error"), caml_copy_string(strerror(errno)));
+  mf->buf = (unsigned char*)malloc(BUFFER_SIZE);
+
+  block = caml_alloc_custom(&madfile_ops, sizeof(madfile_t*), 0, 1);
+  Madfile_val(block) = mf;
+
+  CAMLreturn(block);
+}
+
+CAMLprim value ocaml_mad_openstream(value read_func)
+{
+  CAMLparam1(read_func);
+  CAMLlocal1(block);
+  madfile_t *mf;
+
+  mf = create_mf();
+  mf->read_func = read_func;
+  caml_register_global_root(&mf->read_func);
+  mf->fd = 0;
+  mf->buf = (unsigned char*)malloc(BUFFER_SIZE);
+
+  block = caml_alloc_custom(&madfile_ops, sizeof(madfile_t*), 0, 1);
+  Madfile_val(block) = mf;
+
   CAMLreturn(block);
 }
 
 CAMLprim value ocaml_mad_close(value madf)
 {
   CAMLparam1(madf);
-  madfile_t *mf = madfile_of_block(madf);
-  if (fclose(mf->fd) != 0)
+  madfile_t *mf = Madfile_val(madf);
+  if (!mf->read_func && fclose(mf->fd) != 0)
     caml_raise_with_arg(*caml_named_value("mad_exn_closefile_error"), caml_copy_string(strerror(errno)));
   CAMLreturn(Val_unit);
 }
@@ -149,7 +185,9 @@ CAMLprim value ocaml_mad_close(value madf)
 CAMLprim value ocaml_mad_get_current_position(value madf)
 {
   CAMLparam1(madf);
-  madfile_t *mf = madfile_of_block(madf);
+  madfile_t *mf = Madfile_val(madf);
+  if (mf->read_func)
+    caml_raise_with_arg(*caml_named_value("mad_exn_closefile_error"), caml_copy_string("Not created with openfile."));
   int pos = (int)ftell(mf->fd);
   CAMLreturn(Val_int(pos));
 }
@@ -160,107 +198,114 @@ CAMLprim value ocaml_mad_get_current_position(value madf)
 CAMLprim value ocaml_mad_decode_frame(value madf)
 {
   CAMLparam1(madf);
-  CAMLlocal1(ret);
-  /* TODO: copy *mf instead of using the caml value which could be moved at any moment */
-  madfile_t *mf = madfile_of_block(madf);
+  CAMLlocal2(data, ret);
+  madfile_t *mf = Madfile_val(madf);
   char *output_buf = NULL;
   int output_pos = 0;
-  int i,dec;
+  int i, dec;
 
   /* The input bucket must be filled if it becomes empty
-   */
+  */
   if (mf->stream.buffer == NULL || mf->stream.error == MAD_ERROR_BUFLEN)
+  {
+    size_t read_size, remaining;
+    unsigned char *read_start;
+
+    /* {1} libmad may not consume all bytes of the input
+     * buffer. If the last frame in the buffer is not wholly
+     * contained by it, then that frame's start is pointed by
+     * the next_frame member of the Stream structure. This
+     * common situation occurs when mad_frame_decode() fails,
+     * sets the stream error code to MAD_ERROR_BUFLEN, and
+     * sets the next_frame pointer to a non NULL value. (See
+     * also the comment marked {2} bellow.)
+     *
+     * When this occurs, the remaining unused bytes must be
+     * put back at the beginning of the buffer and taken in
+     * account before refilling the buffer. This means that
+     * the input buffer must be large enough to hold a whole
+     * frame at the highest observable bit-rate (currently 448
+     * kb/s). XXX=XXX Is 2016 bytes the size of the largest
+     * frame? (448000*(1152/32000))/8
+     */
+    if (mf->stream.next_frame)
     {
-      size_t read_size, remaining;
-      unsigned char *read_start;
-
-      /* {1} libmad may not consume all bytes of the input
-       * buffer. If the last frame in the buffer is not wholly
-       * contained by it, then that frame's start is pointed by
-       * the next_frame member of the Stream structure. This
-       * common situation occurs when mad_frame_decode() fails,
-       * sets the stream error code to MAD_ERROR_BUFLEN, and
-       * sets the next_frame pointer to a non NULL value. (See
-       * also the comment marked {2} bellow.)
-       *
-       * When this occurs, the remaining unused bytes must be
-       * put back at the beginning of the buffer and taken in
-       * account before refilling the buffer. This means that
-       * the input buffer must be large enough to hold a whole
-       * frame at the highest observable bit-rate (currently 448
-       * kb/s). XXX=XXX Is 2016 bytes the size of the largest
-       * frame? (448000*(1152/32000))/8
-       */
-      if (mf->stream.next_frame)
-	{
-	  remaining = mf->stream.bufend - mf->stream.next_frame;
-	  memmove(mf->buf, mf->stream.next_frame, remaining);
-	  read_start = mf->buf + remaining;
-	  read_size = BUFFER_SIZE - remaining;
-	}
-      else
-	{
-	  read_size = BUFFER_SIZE;
-	  read_start = mf->buf;
-	  remaining = 0;
-	}
-
-      /* Fill-in the buffer. If an error occurs print a message
-       * and leave the decoding loop. If the end of stream is
-       * reached we also leave the loop but the return status is
-       * left untouched.
-       */
-      // caml_enter_blocking_section();
-      read_size = fread(read_start, 1, read_size, mf->fd);
-      // caml_leave_blocking_section();
-      if (read_size <= 0)
-	{
-	  if(ferror(mf->fd))
-	    {
-	      /* Read error on bitstream. */
-	      caml_raise_with_arg(*caml_named_value("mad_exn_read_error"), caml_copy_string((char*)strerror(errno)));
-	    }
-	  if(feof(mf->fd))
-	    {
-	      /* End of input stream. */
-	      caml_raise_constant(*caml_named_value("mad_exn_end_of_stream"));
-	    }
-	}
-
-      /* {3} When decoding the last frame of a file, it must be
-       * followed by MAD_BUFFER_GUARD zero bytes if one wants to
-       * decode that last frame. When the end of file is
-       * detected we append that quantity of bytes at the end of
-       * the available data. Note that the buffer can't overflow
-       * as the guard size was allocated but not used the the
-       * buffer managment code. (See also the comment marked
-       * {1}.)
-       *
-       * In a message to the mad-dev mailing list on May 29th,
-       * 2001, Rob leslie explains the guard zone as follows:
-       *
-       *    "The reason for MAD_BUFFER_GUARD has to do with the
-       *    way decoding is performed. In Layer III, Huffman
-       *    decoding may inadvertently read a few bytes beyond
-       *    the end of the buffer in the case of certain invalid
-       *    input. This is not detected until after the fact. To
-       *    prevent this from causing problems, and also to
-       *    ensure the next frame's main_data_begin pointer is
-       *    always accessible, MAD requires MAD_BUFFER_GUARD
-       *    (currently 8) bytes to be present in the buffer past
-       *    the end of the current frame in order to decode the
-       *    frame."
-       */
-      /* TODO: pad with MAD_BUFFER_GUARD zero bytes */
-
-      /* Pipe the new buffer content to libmad's stream decoder
-       * facility.
-       */
-      // caml_enter_blocking_section();
-      mad_stream_buffer(&mf->stream, mf->buf, read_size + remaining);
-      // caml_leave_blocking_section();
-      mf->stream.error = 0;
+      remaining = mf->stream.bufend - mf->stream.next_frame;
+      memmove(mf->buf, mf->stream.next_frame, remaining);
+      read_start = mf->buf + remaining;
+      read_size = BUFFER_SIZE - remaining;
     }
+    else
+    {
+      read_size = BUFFER_SIZE;
+      read_start = mf->buf;
+      remaining = 0;
+    }
+
+    /* Fill-in the buffer. If an error occurs print a message
+     * and leave the decoding loop. If the end of stream is
+     * reached we also leave the loop but the return status is
+     * left untouched.
+     */
+    if (mf->read_func)
+    {
+      data = caml_callback(mf->read_func, Val_int(read_size));
+      read_size = Int_val(Field(data, 1));
+      memcpy(read_start, String_val(Field(data, 0)), read_size);
+    }
+    else
+    {
+      read_size = fread(read_start, 1, read_size, mf->fd);
+      if (read_size <= 0)
+      {
+        if(ferror(mf->fd))
+        {
+          /* Read error on bitstream. */
+          caml_raise_with_arg(*caml_named_value("mad_exn_read_error"), caml_copy_string((char*)strerror(errno)));
+        }
+        if(feof(mf->fd))
+        {
+          /* End of input stream. */
+          caml_raise_constant(*caml_named_value("mad_exn_end_of_stream"));
+        }
+      }
+    }
+
+    /* {3} When decoding the last frame of a file, it must be
+     * followed by MAD_BUFFER_GUARD zero bytes if one wants to
+     * decode that last frame. When the end of file is
+     * detected we append that quantity of bytes at the end of
+     * the available data. Note that the buffer can't overflow
+     * as the guard size was allocated but not used the the
+     * buffer managment code. (See also the comment marked
+     * {1}.)
+     *
+     * In a message to the mad-dev mailing list on May 29th,
+     * 2001, Rob leslie explains the guard zone as follows:
+     *
+     *    "The reason for MAD_BUFFER_GUARD has to do with the
+     *    way decoding is performed. In Layer III, Huffman
+     *    decoding may inadvertently read a few bytes beyond
+     *    the end of the buffer in the case of certain invalid
+     *    input. This is not detected until after the fact. To
+     *    prevent this from causing problems, and also to
+     *    ensure the next frame's main_data_begin pointer is
+     *    always accessible, MAD requires MAD_BUFFER_GUARD
+     *    (currently 8) bytes to be present in the buffer past
+     *    the end of the current frame in order to decode the
+     *    frame."
+     */
+    /* TODO: pad with MAD_BUFFER_GUARD zero bytes */
+
+    /* Pipe the new buffer content to libmad's stream decoder
+     * facility.
+     */
+    mad_stream_buffer(&mf->stream, mf->buf, read_size + remaining);
+    mf->stream.error = 0;
+  }
+
+  /***** ENTER BLOCKING SECTION *****/
+  caml_enter_blocking_section();
 
   /* Decode the next mpeg frame. The streams is read from the
    * buffer, its constituents are break down and stored the the
@@ -290,12 +335,13 @@ CAMLprim value ocaml_mad_decode_frame(value madf)
    * this case one can call again mad_frame_decode() in order to
    * skip the faulty part and re-sync to the next frame.
    */
-  // caml_enter_blocking_section();
   dec = mad_frame_decode(&mf->frame, &mf->stream);
-  // caml_leave_blocking_section();
   if (dec)
     if (!MAD_RECOVERABLE(mf->stream.error) && mf->stream.error != MAD_ERROR_BUFLEN)
+    {
+      caml_leave_blocking_section();
       caml_raise_with_arg(*caml_named_value("mad_exn_mad_error"), caml_copy_string(mad_stream_errorstr(&mf->stream)));
+    }
 
   /* Accounting. The computed frame duration is in the frame
    * header structure. It is expressed as a fixed point number
@@ -306,14 +352,15 @@ CAMLprim value ocaml_mad_decode_frame(value madf)
    * some functions of mad's timer module receive some of their
    * mad_timer_t arguments by value!
    */
-  // caml_enter_blocking_section();
   mad_timer_add(&mf->timer, mf->frame.header.duration);
 
   /* Once decoded the frame is synthesized to PCM samples. No errors
    * are reported by mad_synth_frame();
    */
   mad_synth_frame(&mf->synth, &mf->frame);
-  // caml_leave_blocking_section();
+
+  /***** LEAVE BLOCKING SECTION *****/
+  caml_leave_blocking_section();
 
   /* Synthesized samples must be converted from mad's fixed
    * point number to the consumer format. Here we use unsigned
@@ -325,24 +372,24 @@ CAMLprim value ocaml_mad_decode_frame(value madf)
   output_buf = String_val(ret);
 
   for(i = 0; i < mf->synth.pcm.length; i++)
-    {
-      unsigned short sample;
+  {
+    unsigned short sample;
 
-      /* Left channel */
-      sample = ushort_of_madfixed(mf->synth.pcm.samples[0][i]);
-      *(output_buf + output_pos + 0) = sample & 0xff;
-      *(output_buf + output_pos + 1) = sample >> 8;
+    /* Left channel */
+    sample = ushort_of_madfixed(mf->synth.pcm.samples[0][i]);
+    *(output_buf + output_pos + 0) = sample & 0xff;
+    *(output_buf + output_pos + 1) = sample >> 8;
 
-      /* Right channel. If the decoded stream is monophonic then
-       * the right output channel is the same as the left one.
-       */
-      if (MAD_NCHANNELS(&mf->frame.header) == 2)
-	sample = ushort_of_madfixed(mf->synth.pcm.samples[1][i]);
-      *(output_buf + output_pos + 2) = sample & 0xff;
-      *(output_buf + output_pos + 3) = sample >> 8;
+    /* Right channel. If the decoded stream is monophonic then
+     * the right output channel is the same as the left one.
+     */
+    if (MAD_NCHANNELS(&mf->frame.header) == 2)
+      sample = ushort_of_madfixed(mf->synth.pcm.samples[1][i]);
+    *(output_buf + output_pos + 2) = sample & 0xff;
+    *(output_buf + output_pos + 3) = sample >> 8;
 
-      output_pos += 4;
-    }
+    output_pos += 4;
+  }
 
   CAMLreturn(ret);
 }
